package pm.inf.ethz.ch.pancake

import java.io.StringWriter
import scala.language.implicitConversions
import scala.text.Document
import scala.text.Document._
import Syntax._
import Types._

/** A pretty printer for terms and types. Can also pretty-print `Constraint`s
  * and `TypVarSubst`s.
  */
class PrettyPrinter {
  def show(trm: Term, width: Int = 60): String = {
    val doc = toDoc(trm)
    val writer = new StringWriter()
    doc.format(width, writer)

    writer.toString
  }

  def show(ty: Type): String = ty match {
    case Nat => "Nat"
    case Bool => "Bool"
    case TypeVar(id) => id
    case Arrow(ty1, ty2) =>
      val ty1Str = ty1 match {
        case _: Arrow => s"(${show(ty1)})"
        case _ => show(ty1)
      }

      s"$ty1Str -> ${show(ty2)}"
  }

  protected def toDoc(trm: Term): Document = trm match {
    case abs @ Abs(x, t) =>
      val xd = group("lambda" :/: x.id :/: ".")
      val td = toDoc(t)
      gnest(xd :/: td)
    case Ite(t1, t2, t3) =>
      val t1d = gnest("if" :/: toDoc(t1))
      val t2d = gnest("then" :/: toDoc(t2))
      val t3d = gnest("else" :/: toDoc(t3))
      group(t1d :/: t2d :/: t3d)
    case Let(x, t1, t2) =>
      val letd = gnest("let" :/: toDoc(x) :/: "=" :/: toDoc(t1))
      val ind = gnest("in" :/: toDoc(t2))
      group(letd :/: ind)
    case Fix(t) => gnest("fix" :/: toDoc(t))
    case _ => toDocApp(trm)
  }

  protected def toDocApp(trm: Term): Document = trm match {
    case Succ(t) => "succ" :/: toDocAtomic(t)
    case Pred(t) => "pred" :/: toDocAtomic(t)
    case IsZero(t) => "iszero" :/: toDocAtomic(t)
    case App(t1, t2) => group(toDocApp(t1) :/: toDocAtomic(t2))
    case _ => toDocAtomic(trm)
  }

  protected def toDocAtomic(trm: Term): Document = trm match {
    case True => "true"
    case False => "false"
    case NatLit(i) => i
    case Var(id) => id
    case _ => paren(toDoc(trm))
  }

  protected def paren(d: Document) = "(" :: d :: ")"

  protected def gnest(d: Document) = group(nest(2, d))

  protected implicit def string2doc(s: String): Document = text(s)
  protected implicit def int2doc(i: Int): Document = text(i.toString)

  def show(constraints: Constraints): String =
    constraints.map{case (ty1, ty2) => (show(ty1), show(ty2))}
               .map{case (ty1, ty2) => s"$ty1 = $ty2"}
               .mkString(", ")

  def show(typVarSubst: TypVarSubst): String =
    typVarSubst.map{case (tv, typ) => s"${show(tv)} -> ${show(typ)}"}
               .mkString(", ")
}
