package pm.inf.ethz.ch.pancake

import scala.language.implicitConversions

object Syntax {

  /* Abstract syntax tree of our lambda calculus.
   *
   * DO NOT change the signature of the terms!
   * You may, however, add var/val/def declarations to the class body, in
   * which case you might have to change the tree traverser methods defined
   * further down.
   */

  sealed trait Term
  trait BoolLit extends Term
  case object True extends BoolLit
  case object False extends BoolLit
  case class NatLit(n: Int) extends Term
  case class Var(id: String) extends Term
  case class Succ(t: Term) extends Term
  case class Pred(t: Term) extends Term
  case class IsZero(t: Term) extends Term
  case class Abs(x: Var, t: Term) extends Term
  case class App(t1: Term, t2: Term) extends Term
  case class Ite(t1: Term, t2: Term, t3: Term) extends Term
  case class Let(x: Var, t1: Term, t2: Term) extends Term
  case class Fix(t: Term) extends Term

  /* Convenience methods mimicking n-ary abstractions and applications */

  def Abs(x1: Var, x2: Var, xs: List[Var], body: Term): Abs =
    (x1 :: x2 :: xs).foldRight(body){case (x, t) => Abs(x, t)}.asInstanceOf[Abs]

  def Abs(x1: Var, x2: Var, body: Term): Abs =
    Abs(x1, x2, Nil, body)

  def Abs(x1: Var, x2: Var, x3: Var, body: Term): Abs =
    Abs(x1, x2, List(x3), body)

  def Abs(x1: Var, x2: Var, x3: Var, x4: Var, body: Term): Abs =
    Abs(x1, x2, List(x3, x4), body)

  def Abs(x1: Var, x2: Var, x3: Var, x4: Var, x5: Var, body: Term): Abs =
    Abs(x1, x2, List(x3, x4, x5), body)

  def Abs(x1: Var, x2: Var, x3: Var, x4: Var, x5: Var, x6: Var, body: Term): Abs =
    Abs(x1, x2, List(x3, x4, x5, x6), body)

  def App(x: Var, t1: Term, t2: Term, ts: Term*): App =
    ts.foldLeft(App(App(x, t1), t2)){case (app, t) => App(app, t)}

  /* Tree traversal methods. See TraverserTests for some use cases and
   * explanations.
   *
   * Note: Depending on your solution, you might not have to use them.
   */

  /** Returns a list of all direct proper subterms of a given term. For example,
    * the subterms of `Ite(IsZero(Var(""x)), True, False)` are `IsZero(Var("x")),
    * True, False`, that of `IsZero(Var("x"))` is just `Var("x")`, whereas
    * `Var("x")` has no subterms.
    *
    * @param term A terms whose subterms are to be returned
    * @return The list of subterms of `term`
    */
  def subterms(term: Term): List[Term] = term match {
    case _: BoolLit | _: NatLit | _: Var => Nil
    case Succ(t) => List(t)
    case Pred(t) => List(t)
    case IsZero(t) => List(t)
    case Abs(x, t) => List(x, t)
    case App(t1, t2) => List(t1, t2)
    case Ite(t1, t2, t3) => List(t1, t2, t3)
    case Let(x, t1, t2) => List(x, t1, t2)
    case Fix(t) => List(t)
  }

  /** Recursively visits each each subterm `t` of the given `term`, including
    * `term` itself, and applies `f(t)` whenever the partial function `f` is
    * defined for `t`.
    *
    * @param term A term to visit
    * @param f A partial function applied to each subterm of `term`
    */
  def visit(term: Term)(f: PartialFunction[Term, Unit]) {
    if (f.isDefinedAt(term)) f(term)
    for (sub <- subterms(term)) {
      visit(sub)(f)
    }
  }

  /** Transforms a given `term` by visiting each subterm `t` of `term`,
    * including `term` itself, and by replacing `t` by the result of `f(t)`,
    * if the partial function `f` to applicable to `t`.
    *
    * The optional argument `recurseOn` determines whether or not to recurse on
    * subterms. It defaults to only recurse if `f` it undefined for `t`.
    * Passing `_ => true` means recurse on replacements, i.e., results of `f(t)`,
    * as well, whereas `_ => false` prevents any recursion, i.e., `f` is only
    * applied to `term` itself.
    *
    * @param term A terms to transform
    * @param f A partial term substitution function
    * @param recurseOn Determines whether or not to recurse on subterms
    * @return The transformed term
    */
  def transform[T <: Term]
               (term: T)
               (f: PartialFunction[Term, Term] = PartialFunction.empty)
               (recurseOn: Term => Boolean = !f.isDefinedAt(_)): T = {

    def go[X <: Term](root: X): X = transform(root)(f)(recurseOn)

    def recurse(term: Term): Term = term match {
      case _: BoolLit | _: NatLit | _: Var => term
      case Succ(t) => Succ(go(t))
      case Pred(t) => Pred(go(t))
      case IsZero(t) => IsZero(go(t))
      case abs @ Abs(x, t) => Abs(go(x), go(t))
      case App(t1, t2) => App(go(t1), go(t2))
      case Ite(t1, t2, t3) => Ite(go(t1), go(t2), go(t3))
      case Let(x, t1, t2) => Let(go(x), go(t1), go(t2))
      case Fix(t) => Fix(go(t))
    }

    val beforeRecursion = f.applyOrElse(term, identity[Term])

    val afterRecursion =
      if (recurseOn(term)) recurse(beforeRecursion)
      else beforeRecursion

    afterRecursion.asInstanceOf[T]
  }

  /** Creates a deep copy of the given `term`.
    *
    * @param term A term to duplicate
    * @return A deep copy of `term`
    */
  def duplicate(term: Term) = transform(term){case t => t}()
}
