package pm.inf.ethz.ch.pancake

import Syntax._

package object Types {

  /* Types of our lambda calculus.
   *
   * DO NOT change the signature of the types!
   * You may, however, add var/val/def declarations to the class body.
   *
   * Since the types are so few we haven't implemented generic traversers
   * as done terms. See `def typeVariables` further down for an example
   * of a specific traverser over types.
   */

  sealed trait Type
  case object Bool extends Type
  case object Nat extends Type
  case class TypeVar(id: String) extends Type
  case class Arrow(from: Type, to: Type) extends Type

  /* Exceptions related to types and type inference */

  case class NotANaturalNumberException(i: Int)
    extends RuntimeException(s"$i is not a natural number")

  case class UnboundVariableException(v: Var)
    extends RuntimeException(s"Variable $v is not bound")

  case class CircularConstraintException(typ1: TypeVar, typ2: Type)
    extends RuntimeException(s"Circular constraint $typ1 = $typ2")

  case class UnificationException(typ1: Type, typ2: Type)
    extends RuntimeException(s"Cannot unify $typ1 = $typ2")

  /* Some useful types and data structures used across modules */

  type Constraints = List[(Type, Type)]
  def Constraints() = List[(Type, Type)]()

  type TypVarSubst = Map[TypeVar, Type]
  def TypVarSubst() = Map[TypeVar, Type]()

  /** A type inferer. */
  class TypeInferer {

    /* Some useful types and data structures used internally */

    /* The context used during type inference */
    protected type Context /* ??? */
    protected def Context() = ???

    /* A type variable generator produces an infinite stream of type variables
     * with identifiers "T1", "T2", "T3", etc. Call `generator.next()` to get
     * the next fresh type variable.
     */
    protected type TypVarGenerator = Iterator[TypeVar]
    protected def TypVarGenerator(): TypVarGenerator = Stream.from(0).map(i => TypeVar(s"T$i")).iterator

    protected var typVarGen: TypVarGenerator = null
    protected var constraints: Constraints = null

    /** Infers the type of a given `term` by traversing the tree and collecting
      * type constraints.
      *
      * @param term The term to type
      * @return The inferred type and all collected type constraints
      */
    def infer(term: Term): (Type, Constraints) = {
      typVarGen = TypVarGenerator()
      constraints = Constraints()

      val typ = infer(term, Context())
      (typ, constraints)
    }

    /** Performs the actual type inference for `term` in the given `context`.
      * Requires `typVarGen` and `constraints` to be initialised properly.
      * Fresh type variables will be taken from `typVarGen`, and type
      * constraints will be added to `constraints`.
      *
      * @param term The term to type
      * @param context The context to use for typing
      * @return The inferred type
      */
    protected def infer(term: Term, context: Context): Type = ???
      /* Note: It is easier to not use the generic tree traversers defined in
       * Syntax.scala. Just recurse over the terms and follow the type
       * inference rules given in the book.
       */
  }

  /* Functions */

  /** A first-order unification algorithm for type variables.
    *
    * @param constraints A set of type constraints (equalities) that are to be
    *                    unified
    * @return The unified constraints
    */
  def unify(constraints: Constraints): TypVarSubst = {

    /** Performs the actual unification.
      *
      * @param constraints A set of type constraints (equalities) that are to be
      *                    unified
      * @return The unified constraints
      */
    def unify(constraints: Constraints): Constraints = ???
      /* Note: This inner unification method returns constraints again (instead
       * of a TypVarSubst) because the safety checks further down wouldn't make
       * sense otherwise. The checks are only there to point out problems that
       * an incorrect implementation of the unification algorithm is likely to
       * have.
       * However, you are free to remove the checks and to change the signature
       * of the *inner* unify-method, as long as you don't change the signature
       * of the *outer* method.
       */

    /* Unify the constraints by calling the inner unification method */
    val unifiedConstraints = unify(constraints)

    /* Some safety checks */

    unifiedConstraints.foreach{
      case (_: TypeVar, _) => /* ok */
      case (typ1, typ2) => sys.error(s"Expected unified constraints to replace type variables, but found $typ1 |-> $typ2")
    }

    unifiedConstraints.map(_._1).diff(unifiedConstraints.map(_._1).distinct) match {
      case Seq() => /* ok */
      case tvs => sys.error("Type variables $tvs are ambiguously substituted by the allegedly unified constraints $unifiedConstraints")
    }

    /* Looks ok, we just have to convince the type system that the constraints
     * form a map from type variables to types.
     */

    (TypVarSubst() ++ unifiedConstraints).asInstanceOf[TypVarSubst]
  }

  /* Substitution functions */

  /** Substitutes `orig` in `in` with `repl`. If `in` equals `orig` then the
    * result is `repl`.
    *
    * @param orig A type to replace
    * @param in A type in which `orig` is to be replaced by `repl`
    * @param repl The substitute for `orig`
    * @return A version of `in` where each occurrence of `orig` has been
    *         substituted with `repl`
    */
  def subst(orig: Type, in: Type, repl: Type): Type = in match {
    case tv: TypeVar =>
      if (tv == orig) repl
      else tv
    case Arrow(typ1, typ2) =>
      Arrow(subst(orig, typ1, repl), subst(orig, typ2, repl))
    case _ => in
  }

  /** Lifts `subst(Type, Type, Type): Type` to constraints by substituting
    * `orig` in each pair `(typ1, typ2)` in `in` with `repl`.
    *
    * @param orig A type to replace
    * @param in A list of constraints in which each occurrence of `orig` is to
    *           be replaced by `repl`
    * @param repl The substitute for `orig`
    * @return A version of `in` where each occurrence of `orig` has been
    *         substituted with `repl`
    */
  def subst(orig: Type, in: Constraints, repl: Type): Constraints = ???

  /** Applies the type variable substitution `typVarSubst` to `in`.
    * Following the book, the substitution is not transitive. That is, if a
    * substitution mapping `X -> Y` and `Y -> Nat` is applied to `X -> Y`,
    * the result will be `Y -> Nat` and not `Nat -> Nat`.
    *
    * @param typVarSubst The substitution to apply to `in`
    * @param in A type to perform substitution on
    * @return The result of applying `typVarSubst` to `in`
    */
  def applySubstitution(typVarSubst: TypVarSubst, in: Type): Type = ???

  /** A transitive version of `applySubstitution(TypVarSubst, Type): Type`.
    * That is, if a substitution mapping `X -> Y` and `Y -> Nat` is applied to
    * `X -> Y`, the result will be `Nat -> Nat`.
    *
    * Attention: The behaviour is undefined for substitutions that can be
    * applied indefinitely, e.g., `X -> X`.
    *
    * @param typVarSubst The substitution to apply to `in`
    * @param in A type to perform substitution on
    * @return The result of transitively applying `typVarSubst` to `in`
    */
  def applySubstitutionTransitively(typVarSubst: TypVarSubst, in: Type): Type = ???

  /** The list of type variables occurring in a given `typ`.
    *
    * @param typ A type
    * @return The list of type variables occurring in `typ`
    */
  def typeVariables(typ: Type): List[TypeVar] = typ match {
    case tv: TypeVar => tv :: Nil
    case Arrow(typ1, typ2) => typeVariables(typ1) ::: typeVariables(typ2)
    case Bool | Nat => Nil
  }

  /** Returns a type variable substitution that maps the first type variable
    * in `typVars` to the type variable "A", the second to "B", the third to
    * "C", etc. * Throws an exception if more than 26 type variables are given.
    *
    * @param typVars A list of type variables (at most 26)
    * @return A mapping from the list of type variables to type variables "A",
    *         "B", "C", etc,
    */
  def prettyTypVarSubst(typVars: List[TypeVar]): TypVarSubst = {
    assert(typVars.length <= 26, "prettyTypVarSubst cannot handle more than 26 type variables")

    val azGen = ('A' to 'Z').iterator.map(c => TypeVar(c.toString))
    typVars.map((_, azGen.next())).toMap
  }
}
