package pm.inf.ethz.ch.pancake

// import Predef.{ArrowAssoc => _, any2ArrowAssoc => _, _}
  /* Predef defines an arrow operator as well. It doesn't seem necessary to
   * "unimport" it because it is shadowed by ours imported from DSL, but let's
   * play it safe and keep the corresponding import around.
   */
import org.scalatest.{Matchers, FunSuite}
import Syntax._
import Types._
import DSL._

class TermTests extends FunSuite with Matchers {
  val typeInferer = new TypeInferer

  def inferType(term: Term): Type = {
    /* Infer the type of term */
    val (typ, constraints) = typeInferer.infer(term)
    /* Unify set of constraints to get a type variable substitution */
    val typVarSubst = unify(constraints)
    /* Transitively apply the computed type variable substitution */
    val unifiedTyp = applySubstitutionTransitively(typVarSubst, typ)
    /* Get all remaining type variables */
    val typeVars = typeVariables(unifiedTyp).distinct

    /* Apply a substitution to the term that renames the remaining type
     * variables to A, B, C, etc.
     */
    applySubstitution(prettyTypVarSubst(typeVars), unifiedTyp)
  }

  ignore("not a natural number") {
    a [NotANaturalNumberException] should be thrownBy
      inferType(-1)

    a [NotANaturalNumberException] should be thrownBy
      inferType(Let(x, -10, 0))
  }

  ignore("undeclared variables") {
    a [UnboundVariableException] should be thrownBy
      inferType(App(x, 0))

    a [UnboundVariableException] should be thrownBy
      inferType(App(0, x))

    a [UnboundVariableException] should be thrownBy
      inferType(Let(x, 0, Pred(y)))
  }

  ignore("unification impossible") {
    a [UnificationException] should be thrownBy
      inferType(App(Succ(0), true))

    a [UnificationException] should be thrownBy
      inferType(Abs(x, Ite(IsZero(x), x, IsZero(x))))
  }

  ignore("very simple terms") {
    inferType(true) shouldBe Bool
    inferType(8) shouldBe Nat
    inferType(IsZero(0)) shouldBe Bool
    inferType(Pred(Succ(1))) shouldBe Nat
  }

  ignore("slightly more involved terms") {
    inferType(App(Abs(x, Succ(x)), 0)) shouldBe Nat
    inferType(Abs(x, y, Ite(IsZero(x), y, x))) shouldBe Nat → (Nat → Nat)
    inferType(Abs(f, App(f, 1))) shouldBe (Nat → A) → A
    inferType(Abs(f, App(f, 1, 0))) shouldBe (Nat → (Nat → A)) → A
    inferType(Abs(f, x, y, App(f, x, y))) shouldBe (A → (B → C)) → (A → (B → C))
  }

  ignore("let-expressions") {
    inferType(Let(x, 0, x)) shouldBe Nat
    inferType(Let(f, Abs(f, IsZero(f)), App(f, 0))) shouldBe Bool
  }

  ignore("recursion") {
    val add_core =
      Abs(r, x, y,
        Ite(IsZero(x),
            y,
            App(r, Pred(x), Succ(y))))

    inferType(add_core) shouldBe (Nat → (Nat → Nat)) → (Nat → (Nat → Nat))

    inferType(Fix(add_core)) shouldBe (Nat → (Nat → Nat))
  }

  ignore("circular constraints") {
    a [CircularConstraintException] should be thrownBy
      inferType(
        Abs(f,
            App(Abs(x, App(f, App(x, x))),
                Abs(x, App(f, App(x, x))))))
  }

  ignore("identity") {
    val idt = Abs(x, x)

    inferType(idt) shouldBe A → A
    inferType(App(idt, idt)) shouldBe A → A
    inferType(Let(id, idt, App(id, 0))) shouldBe Nat

    inferType(
      Let(id,
         idt,
         App(Abs("_", App(id, true)),
             App(id, 0)))
    ) shouldBe Bool

    inferType(
      Abs(b, x,
          Let(id,
              idt,
              Ite(App(id, b), Succ(x), App(id, x))))
    ) shouldBe Bool → (Nat → Nat)

    val ft = Abs(f, x, Ite(true, idt, x))
    val inct = Abs(x, Succ(x))

    inferType(
      Let(f, ft,
      Let(id, idt,
      Let(inc, inct,
          Ite(true, App(f, id), App(f, inct)))))
    ) shouldBe (A → A) → (A → A)
  }
}
