package pm.inf.ethz.ch.pancake

import org.scalatest.{Matchers, FunSuite}
import Syntax._
import DSL._

class TraverserTests extends FunSuite with Matchers {
  val term1 = Ite(IsZero(x), true, false)
  val term2 = Abs(x, y, Ite(IsZero(x), IsZero(y), false))

  test("visit terms") {
    val terms = collection.mutable.ListBuffer[Term]()

    visit(term1){case b: BoolLit => terms += b}
    terms shouldBe List(True, False)
    terms shouldBe List[Term](true, false)
      /* Note: The test would fail if List(true, false) where used because
       * Scala would not apply the implicit conversions from true/false to
       * True/False, since having a List[Bool] is fine here because shouldBe
       * is not strongly typed (it takes an Any). Thus, we have to make the
       * need for conversion explicit by writing List[Term](true, false).
       */
    terms.clear()

    visit(term2){case abs: Abs => terms += abs}
    terms shouldBe List(Abs(x, y, Ite(IsZero(x), IsZero(y), false)),
                        Abs(y, Ite(IsZero(x), IsZero(y), false)))
    terms.clear()

    visit(term2) {
      case v @ Var("x") => terms += v
      case v @ Var(name) if name == "y" => terms += v
      case IsZero(v) => terms += v
    }
    terms shouldBe List(x, y, x, x, y, y)
    terms.clear()
  }

  test("transform terms") {
    var tt: Term = null

    tt = transform(Ite(True, x, y)){case Ite(True, t1, _) => t1}()
    tt shouldBe x

    tt = transform(term1){
      case True => False
        /* Note: We cannot just use true in the match pattern because the partial
         * function then be typed as Boolean -> Term, which is not the type
         * transform expects. Additional type annotations would be needed.
         */
    }()
    tt shouldBe Ite(IsZero(x), false, false)

    tt = transform(term1){case IsZero(v) => IsZero(Pred(v))}()
    tt shouldBe Ite(IsZero(Pred(x)), true, false)

    tt = transform(term1){
      case IsZero(`x`) => IsZero(Pred(x))
      case `x` => y
      /* Note: We cannot use x as the match pattern because Scala assumes
       * that patterns starting with a lower-case character are local
       * variables that are be bound in the match case, and that upper-case
       * patterns are elements, e.g., case classes, that are to be matched on.
       */
    }(_ => true)
      /* Note: Passing _ => true tells transform to apply the transformation
       * to the result of earlier matches as well. Here, the second case matches
       * the x occurring inside IsZero(Pred(x)).
       * The default behaviour of transform is to not recurse on subterms of a
       * term that already matched. That is, the result of this transformation
       * would be Ite(IsZero(Pred(x)), true, false) because variable x would
       * not have been replaced by y.
       */
    tt shouldBe Ite(IsZero(Pred(y)), true, false)
  }
}
